/**
 * jdasm.c -- Deassembler implementation file.
 *
 * Author: Gaurav Vaish <gvaish@iitk.ac.in>
 *         Gopal V (gopalv82@yahoo.com)
 *
 * (C) Gaurav Vaish, Gopal V; 2002
 */

#include <jdasm.h>
#include <jopcodes.h>
#include <jparser.h>

void print_interfaces(JVClass *klass, JVCPEntry *cp_entry)
{
	int i;
	for(i = 0; i < klass->interface_count; i++)
	{
		printf(".implements %s\n", klass->interface_entries[i].full_name);
	}
}

void print_fields(JVClass *klass, JVCPEntry *cp_entry)
{
	int i;
	for(i = 0; i < klass->field_count; i++)
	{
		printf(".field %s %s %s\n", klass->field_entries[i].full_flags, \
			klass->field_entries[i].full_name, klass->field_entries[i].full_type);
	}
}

void print_string_cp(JVCPEntry *cp_entry, int index)
{
	switch(cp_entry[index].type)
	{
		case UTF8:     printf("%s", cp_entry[index].values.utf8string);
		               break;
		case STRING:   printf("\"%s\"", cp_entry[cp_entry[index].values.strvalue].values.utf8string);
		               break;
		case INTEGER:  printf("%d", cp_entry[index].values.intvalue);
		               break;
		case LONG:     printf("%ld", cp_entry[index].values.longvalue);
		               break;
		case FLOAT:    printf("%f", cp_entry[index].values.floatvalue);
		               break;
		case DOUBLE:   printf("%d", cp_entry[index].values.doublevalue);
		               break;
		case CLASSREF: printf("%s", cp_entry[index].values.class_value.full_name);
		               break;
		case FIELDREF: printf("%s/%s %s", cp_entry[index].values.ref_value.full_class, \
		                   cp_entry[index].values.ref_value.full_name, \
		                   cp_entry[index].values.ref_value.full_type);
		               break;
		case METHODREF:
		case INTERFACEMETHODREF:
		               printf("%s/%s %s", cp_entry[index].values.ref_value.full_class, \
		                   cp_entry[index].values.ref_value.full_name, \
		                   cp_entry[index].values.ref_value.full_type);
		               break;
		case NAMEANDTYPE:
		               printf("%s", resolve_name_type(index, cp_entry));
		               break;
	}
}

void print_byte_code(byte *code, int *start_index, int code_len, JVCPEntry *cp_entry)
{
	int   start = *start_index;
	Int32 key;
	Int32 offset;
	Int32 default_offset;
	Int32 count;
	int   insn_offset = start;
	int   j;

	int   i = search_opcode_map(code[start]);
	opcode_map found;

	if(i >= 0)
	{
		found = get_opcode_map(i);
		switch(found.args)
		{
			case NONE:   PRINT_OP_NAME(found);
			             start++;
			             break;
			case I:      PRINT_OP_NAME(found);
			             start += 2;
			             break;
			case II:     PRINT_OP_NAME(found);
			             start += 3;
			             break;
			case FIELD:
			case METHOD:
			case CLASS:
			case BIGCONSTANT:
			             PRINT_OP_NAME(found);
			             print_string_cp(cp_entry, (code[start + 1] << 0x08) | code[start + 2]);
			             start += 3;
			             break;
			case CONSTANT:
			             PRINT_OP_NAME(found);
			             print_string_cp(cp_entry, code[start + 1]);
			             start += 2;
			             break;
			case SWITCH: PRINT_OP_NAME(found);
			             start++;
			             while( (start % 4) != 0)
			             {
							 start++;
					     }
					     default_offset = (((Int32)code[start]) << 0x18) | \
					                      (((Int32)code[start + 1]) << 0x10) | \
					                      (((Int32)code[start + 2]) << 0x08) | \
					                      ((Int32)code[start + 3]);
					     start += 4;
					     count = (((Int32)code[start]) << 0x18) | \
					             (((Int32)code[start + 1]) << 0x10) | \
					             (((Int32)code[start + 2]) << 0x08) | \
					             ((Int32)code[start + 3]);
					     start += 4;
					     printf(" %ld key-values\n", count);
					     for(j = 0; j < count; j++)
					     {
						     key = (((Int32)code[start]) << 0x18) | \
					               (((Int32)code[start + 1]) << 0x10) | \
					               (((Int32)code[start + 2]) << 0x08) | \
					               ((Int32)code[start + 3]);
					         start += 4;
					         offset = (((Int32)code[start]) << 0x18) | \
					                  (((Int32)code[start + 1]) << 0x10) | \
					                  (((Int32)code[start + 2]) << 0x08) | \
					                  ((Int32)code[start + 3]);
					         start += 4;
					         printf("\t\t%ld\t: %ld\n", key, insn_offset + offset);
					     }
					     printf("\t\tdefault: %ld\n", insn_offset + default_offset);
					     break;
			case LABEL:  PRINT_OP_NAME(found);
			             printf("\t %s;code offset ", insn_offset + \
			                                          ( ((Int16)code[start + 1]) << 0x08 | \
			                                            ((Int16)code[start + 2]) ) );
			             start += 3;
			             break;
			default:     printf(": Unknown instruction %d", code[start]);
			             start++;
		}
		printf("\n");
	}
	*start_index = start;
}

void print_method_code(JVMethodEntry *method, JVCPEntry *cp_entry)
{
	int   i;
	int   start   = 0;
	byte *code    = method->attr_entries[method->code_index].attributes.code.code;
	int   codelen = method->attr_entries[method->code_index].attributes.code.code_count;

	printf(".limit stack %d\n", method->attr_entries[method->code_index].attributes.code.max_stack);
	printf(".limit locals %d\n", method->attr_entries[method->code_index].attributes.code.max_locals);

	if( (method->acc_flags & ACCESS_ABSTRACT) != 0 || codelen == 0)
	{
		return;
	}

	while(start < codelen)
	{
		print_byte_code(code, &start, codelen, cp_entry);
	}
	printf(": TODO-> label management\n");
	printf(": TODO-> wide opcode not supported\n");
}

void print_methods(JVClass *klass, JVCPEntry *cp_entry)
{
	int i;
	for(i = 0; i < klass->method_count; i++)
	{
		printf(".method %s %s%s\n", klass->method_entries[i].full_flags, \
		                            klass->method_entries[i].full_name,
		                            klass->method_entries[i].full_type);
		print_method_code(&((klass->method_entries)[i]), cp_entry);
		printf(".end method\n");
	}
}

void print_class(JVClass *klass, JVCPEntry *cp_entry)
{
	printf(".class %s %s\n", klass->full_flags, klass->full_name);
	printf(".super %s\n", klass->full_super);
	print_interfaces(klass, cp_entry);
	print_fields(klass, cp_entry);
	print_methods(klass, cp_entry);
}

void usage()
{
	printf("\tjdasm 0.0.1 - Java Deassembler\n");
	printf("\n");
	printf("usage: jdasm [OPTIONS] <class-file>\n\n");
	printf("Options:\n");
	printf("\t--version\n\t  Print version and exit\n");
	printf("\t--help\n\t  Print this help screen\n");
}

void version()
{
	printf("\tjdasm 0.0.1 - Java Deassembler\n");
	printf("\n");
}

int main(int argc, char **argv)
{
	FILE *fp;
	JVClass *klass;
	if(argc <= 1)
	{
		usage();
		return;
	}
	while(argv[1][0] == '-' && argc > 1)
	{
		if(strcmp(argv[1], "--version") == 0)
		{
			version();
			return 0;
		} else if(strcmp(argv[1], "--help") == 0)
		{
			usage();
			return 0;
		} else
		{
			fprintf(stderr, "Unknown option: %s\n", argv[1]);
			fprintf(stderr, "Try --help for valid options.\n");
			return 1;
		}
		argv++;
		argc--;
	}
	fp = fopen(argv[1], "rb");
	if(fp == NULL)
	{
		fprintf(stderr, "Cannot open file: %s\n", argv[1]);
		return 2;
	}
	klass = (JVClass *)calloc(1, sizeof(JVClass));
	if(read_class(fp, klass) != OK)
	{
		fprintf(stderr, "Cannot read file: %s\n", argv[1]);
		fclose(fp);
		return 3;
	}
	resolve_class(klass, klass->cp_entries);
	print_class(klass, klass->cp_entries);
	return 0;
}