/**
 * jparser.c -- Parser source file.
 *       All the parsing of the class file is done here.
 *
 * Author: Gaurav Vaish <gvaish@iitk.ac.in>
 *         Gopal V (gopalv82@yahoo.com)
 *
 * (C) Gaurav Vaish, Gopal V; 2002
 */

#include <jparser.h>

/**
 * Contant Pool Entries Table : Literal Strings
 */

char* CPTable[] = {
	"Junk",
	"CONSTANT_Utf8",
	"Junk",
	"CONSTANT_Integer",
	"CONSTANT_Float",
	"CONSTANT_Long",
	"CONSTANT_Double",
	"CONSTANT_Class",
	"CONSTANT_String",
	"CONSTANT_Field",
	"CONSTANT_Method",
	"CONSTANT_InterfaceMethod",
	"CONSTANT_NameAndType",
	"Junk",
	"Junk",
	"Junk"
};

/**
 * File reader methods
 */

inline UInt8 read_byte(FILE *fp)
{
	return fgetc(fp);
}

inline UInt16 read_uint16(FILE *fp)
{
	return ( (((UInt16)read_byte(fp)) << 0x08) | ((UInt16)read_byte(fp)) );
}

inline UInt32 read_uint32(FILE *fp)
{
	return ( (((UInt32)read_uint16(fp)) << 0x10) | ((UInt32)read_uint16(fp)) );
}

inline UInt64 read_uint64(FILE *fp)
{
	return ( (((UInt64)read_uint32(fp)) << 0x20) | ((UInt64)read_uint32(fp)) );
}

inline void skip_bytes(FILE *fp, int count)
{
	int i;
	for(i = 0; i < count; i++)
	{
		fgetc(fp);
	}
}

inline byte buffer_read_byte(jbuffer *buffer)
{
	byte retVal;
	if(buffer->left == 0)
	{
		//fprintf(stderr, "Empty buffer\n");
		return 0;
	}

	(buffer->left)--;
	retVal = buffer->data[buffer->next];
	(buffer->next)++;
	return retVal;
}

inline UInt16 buffer_read_uint16(jbuffer *buffer)
{
	return ( (((UInt16)buffer_read_byte(buffer)) << 0x08) | ((UInt16)buffer_read_byte(buffer)) );
}

inline UInt32 buffer_read_uint32(jbuffer *buffer)
{
	return ( (((UInt32)buffer_read_uint16(buffer)) << 0x10) | ((UInt32)buffer_read_uint16(buffer)) );
}

void buffer_skip_bytes(jbuffer *buffer, UInt32 count)
{
	if(buffer->left < count || count < 0 )
	{
		// fprintf(stderr, "Illegal call");
		return;
	}
	(buffer->next) += count;
	(buffer->left) -= count;
}

int read_cp_entry(FILE *fp, JVCPEntry *cp_entry)
{
	UInt32 ui32;
	UInt64 ui64;
	cp_entry->type = read_byte(fp);
	switch(cp_entry->type)
	{
		case UTF8:    cp_entry->length = read_uint16(fp);
		              cp_entry->values.utf8string = (char *)calloc(cp_entry->length + 1, sizeof(char));
		              fread(cp_entry->values.utf8string, cp_entry->length, sizeof(char), fp);
		              break;
		case INTEGER: cp_entry->values.intvalue = read_uint32(fp);
		              break;
		case FLOAT:   ui32 = read_uint32(fp);
		              memcpy(&(cp_entry->values.floatvalue), &ui32, sizeof(float));
		              break;
		case LONG:    cp_entry->values.longvalue = read_uint64(fp);
		              cp_entry->skip = 1;
		              break;
		case DOUBLE:  ui64 = read_uint64(fp);
		              cp_entry->skip = 1;
		              memcpy(&(cp_entry->values.doublevalue), &ui64, sizeof(float));
		              break;
		case STRING:  cp_entry->values.strvalue = read_uint16(fp);
		              break;
		case CLASSREF:
		              cp_entry->values.class_value.name_index = read_uint16(fp);
		              break;
		case FIELDREF:
		case METHODREF:
		case INTERFACEMETHODREF:
		              cp_entry->values.ref_value.class_index = read_uint16(fp);
		              cp_entry->values.ref_value.name_type   = read_uint16(fp);
		              break;
		case NAMEANDTYPE:
		              cp_entry->values.name_type.name_index = read_uint16(fp);
		              cp_entry->values.name_type.type_index = read_uint16(fp);
		              break;
		default:      cp_entry->type = 0;
                    // return ERROR;
	}
	return OK;
}

int read_cp(FILE *fp, JVClass *klass)
{
	int i;
	klass->cp_count = read_uint16(fp);
	klass->cp_entries = (JVCPEntry *)calloc(klass->cp_count, sizeof(JVCPEntry));

	for(i = 1; i < klass->cp_count; i++)
	{
		read_cp_entry(fp, &klass->cp_entries[i]);
		i += (klass->cp_entries)[i].skip;
	}
	return OK;
}

int read_flags(FILE *fp, JVClass *klass)
{
	klass->acc_flags = read_uint16(fp);
	klass->this_class = read_uint16(fp);
	klass->super_class = read_uint16(fp);
	return OK;
}

int read_interfaces(FILE *fp, JVClass *klass)
{
	int i;
	klass->interface_count = read_uint16(fp);
	klass->interface_entries = (JVInterfaceEntry *)calloc(klass->interface_count, sizeof(JVInterfaceEntry));

	for(i = 0; i < klass->interface_count; i++)
	{
		klass->interface_entries[i].index = read_uint16(fp);
	}
	return OK;
}

JVAttributeEntry* read_attributes(FILE *fp, int count, int type)
{
	int i;
	JVAttributeEntry *retVal = (JVAttributeEntry *)calloc(count, sizeof(JVAttributeEntry));

	for(i = 0; i < count; i++)
	{
		retVal[i].type = type;
		retVal[i].name_index = read_uint16(fp);
		retVal[i].byte_count = read_uint32(fp);
		retVal[i].buffer     = (jbuffer *)calloc(1, sizeof(jbuffer));
		retVal[i].buffer->data = (byte *)calloc(retVal[i].byte_count, sizeof(byte));
		retVal[i].buffer->left = retVal[i].byte_count;
		fread(retVal[i].buffer->data, sizeof(byte), retVal[i].byte_count, fp);
	}
	return retVal;
}

int read_fields(FILE *fp, JVClass *klass)
{
	int i;
	klass->field_count = read_uint16(fp);
	klass->field_entries = (JVFieldEntry *)calloc(klass->field_count, sizeof(JVFieldEntry));

	for(i = 0; i < klass->field_count; i++)
	{
		klass->field_entries[i].acc_flags  = read_uint16(fp);
		klass->field_entries[i].name_index = read_uint16(fp);
		klass->field_entries[i].type_index = read_uint16(fp);
		klass->field_entries[i].attr_count = read_uint16(fp);
		klass->field_entries[i].attr_entries = read_attributes(fp, klass->field_entries[i].attr_count, ATTRIBUTE_FIELD);
	}
	return OK;
}

int read_methods(FILE *fp, JVClass *klass)
{
	int i;
	klass->method_count = read_uint16(fp);
	klass->method_entries = (JVMethodEntry *)calloc(klass->method_count, sizeof(JVMethodEntry));

	for(i = 0; i < klass->method_count; i++)
	{
		klass->method_entries[i].acc_flags  = read_uint16(fp);
		klass->method_entries[i].name_index = read_uint16(fp);
		klass->method_entries[i].type_index = read_uint16(fp);
		klass->method_entries[i].attr_count = read_uint16(fp);
		klass->method_entries[i].attr_entries = read_attributes(fp, klass->method_entries[i].attr_count, ATTRIBUTE_METHOD);
	}
	return OK;
}

int read_class(FILE *fp, JVClass *klass)
{
	int errors = OK;
	klass->magic = read_uint32(fp);
	if(klass->magic != 0xCAFEBABE)
	{
		fprintf(stderr, "Magic number does not match.\n");
		fprintf(stderr, "Not a valid class file.\n");
		return ERROR;
	}
	klass->version_minor = read_uint16(fp);
	klass->version_major = read_uint16(fp);

	errors |= read_cp(fp, klass);
	read_flags(fp, klass);
	read_interfaces(fp, klass);

	read_fields(fp, klass);
	read_methods(fp, klass);
	klass->attr_count = read_uint16(fp);
	klass->attr_entries = read_attributes(fp, klass->attr_count, ATTRIBUTE_TOP);


	return errors;
}

/**
 * Resolving methods (resolve all entries in the class)
 */

char* resolve_utf8(UInt16 index, JVCPEntry *cp_entry)
{
	/*
	if(cp_entry[index].type!= UTF8)
	{
		fprintf(stderr,"OOPS !\n");
	}
	*/
	return cp_entry[index].values.utf8string;
}

char* resolve_string(UInt16 index, JVCPEntry *cp_entry)
{
	return resolve_utf8(cp_entry[index].values.strvalue, cp_entry);
}

char* resolve_classref(UInt16 index, JVCPEntry *cp_entry)
{
	cp_entry[index].values.class_value.full_name = \
		resolve_utf8(cp_entry[index].values.class_value.name_index, cp_entry);
	return cp_entry[index].values.class_value.full_name;
}

JVCPEntry* resolve_name_type(UInt16 index, JVCPEntry *cp_entry)
{
	JVCPEntry *retVal = &(cp_entry[index]);

	retVal->values.name_type.full_name = \
		resolve_utf8(retVal->values.name_type.name_index, cp_entry);
	retVal->values.name_type.full_type = \
		resolve_utf8(retVal->values.name_type.type_index, cp_entry);

	return retVal;
}

JVCPEntry* resolve_ref(UInt16 index, JVCPEntry *cp_entry)
{
	JVCPEntry *ref = &(cp_entry[index]);
	JVCPEntry *name_type = resolve_name_type(ref->values.ref_value.name_type,cp_entry);

	ref->values.ref_value.full_class = \
		resolve_classref(ref->values.ref_value.class_index, cp_entry);
	ref->values.ref_value.full_name  = \
		name_type->values.name_type.full_name;
	ref->values.ref_value.full_type  = \
		name_type->values.name_type.full_type;

	return ref;
}

int resolve_cp(JVClass *klass, JVCPEntry *cp_entry)
{
	int i;
	for(i = 0; i < klass->cp_count; i++)
	{
		switch(cp_entry[i].type)
		{
			case CLASSREF:  resolve_classref(i, cp_entry);
			                break;
			case FIELDREF:
			case METHODREF:
			case INTERFACEMETHODREF:
			                resolve_ref(i, cp_entry);
			                break;
			case NAMEANDTYPE:
			                resolve_name_type(i, cp_entry);
			                break;
		}
	}
	return OK;
}

int is_set(UInt16 value, UInt16 bits)
{
	return ( ((value & bits) != 0x00) ? 1 : 0 );
}

char* resolve_flags(UInt16 flags)
{
	char *retVal = (char *)calloc(101, sizeof(char));

	if(flags == 0x0000)
		strcat(retVal, " ");
	if(is_set(flags, ACCESS_PUBLIC))
		strcat(retVal, "public ");
	if(is_set(flags, ACCESS_PRIVATE))
		strcat(retVal, "private ");
	if(is_set(flags, ACCESS_PROTECTED))
		strcat(retVal, "protected ");
	if(is_set(flags, ACCESS_STATIC))
		strcat(retVal, " static ");
	if(is_set(flags, ACCESS_FINAL))
		strcat(retVal, " final ");
	if(is_set(flags, ACCESS_SYNCHRONIZED))
		strcat(retVal, " synchronized ");
	if(is_set(flags, ACCESS_VOLATILE))
		strcat(retVal, " volatile ");
	if(is_set(flags, ACCESS_TRANSIENT))
		strcat(retVal, " transient ");
	if(is_set(flags, ACCESS_NATIVE))
		strcat(retVal, " native ");
	if(is_set(flags, ACCESS_ABSTRACT))
		strcat(retVal, " abstract ");

	retVal = (char *)realloc(retVal, strlen(retVal));

	return retVal;
}

int resolve_all_interfaces(JVClass *klass, JVCPEntry *cp_entry)
{
	int i;
	for(i = 0; i < klass->interface_count; i++)
	{
		klass->interface_entries[i].full_name = \
			resolve_classref(klass->interface_entries[i].index, cp_entry);
	}

	return OK;
}

int resolve_top_attributes(JVClass *klass, JVCPEntry *cp_entry)
{
	int retVal = OK;
	int i;
	for(i = 0; i < klass->attr_count; i++)
	{
		klass->attr_entries[i].full_name = \
			resolve_utf8(klass->attr_entries[i].name_index, cp_entry);
		retVal |= process_source_attribute(&((klass->attr_entries)[i]), cp_entry);
	}
	return retVal;
}

int resolve_field(JVFieldEntry *field_entry, JVCPEntry *cp_entry)
{
	field_entry->full_flags = resolve_flags(field_entry->acc_flags);
	field_entry->full_name  = resolve_utf8(field_entry->name_index, cp_entry);
	field_entry->full_type  = resolve_utf8(field_entry->type_index, cp_entry);

	return OK;
}

int resolve_all_fields(JVClass *klass,JVCPEntry *cp_entry)
{
	int errors = OK;
	int i;
	for(i = 0; i < klass->field_count; i++)
	{
		resolve_field(&((klass->field_entries)[i]), cp_entry);
	}
	return OK;
}

int resolve_all_method_attributes(JVMethodEntry *method,JVCPEntry *cp_entry)
{
	int i;
	for(i = 0; i < method->attr_count; i++)
	{
		method->attr_entries[i].full_name = \
			resolve_utf8(method->attr_entries[i].name_index, cp_entry);
		if(strncmp("Code", method->attr_entries[i].full_name, 4) == 0)
		{
			method->code_index = i;
			process_methodcode_attribute(&((method->attr_entries)[i]), cp_entry);
		}
	}
	return OK;
}

int resolve_method(JVMethodEntry *method,JVCPEntry *cp_entry)
{
	method->full_flags = resolve_flags(method->acc_flags);
	method->full_name  = resolve_utf8(method->name_index, cp_entry);
	method->full_type  = resolve_utf8(method->type_index, cp_entry);
	resolve_all_method_attributes(method, cp_entry);

	return OK;
}

int resolve_all_methods(JVClass *klass,JVCPEntry *cp_entry)
{
	int i;
	for(i = 0; i < klass->method_count; i++)
	{
		resolve_method(&((klass->method_entries)[i]), cp_entry);
	}

	return OK;
}

int resolve_class(JVClass *klass,JVCPEntry *cp_entry)
{
	klass->full_flags = resolve_flags(klass->acc_flags);
	klass->full_name  = resolve_classref(klass->this_class, cp_entry);
	klass->full_super = resolve_classref(klass->super_class, cp_entry);

	resolve_cp(klass, cp_entry);
	resolve_all_interfaces(klass, cp_entry);
	resolve_all_fields(klass, cp_entry);
	resolve_all_methods(klass, cp_entry);
	resolve_top_attributes(klass, cp_entry);

	return OK;
}

int process_source_attribute(JVAttributeEntry *attr_entry, JVCPEntry *cp_entry)
{
	if(attr_entry->type == ATTRIBUTE_TOP && attr_entry->byte_count == 2)
	{
		if(strncmp("SourceFile", resolve_utf8(attr_entry->name_index, cp_entry), 10) == 0)
		{
			attr_entry->attributes.source.source_index = buffer_read_uint16(attr_entry->buffer);
			attr_entry->attributes.source.full_source  = \
				resolve_utf8(attr_entry->attributes.source.source_index, cp_entry);
			printf(".source %s\n", attr_entry->attributes.source.full_source);
		}
		//return OK;
	}
	//return ERROR;
	return OK;
}

int process_methodcode_attribute(JVAttributeEntry *code, JVCPEntry *cp_entry)
{
	int i;
	code->attributes.code.max_stack  = buffer_read_uint16(code->buffer);
	code->attributes.code.max_locals = buffer_read_uint16(code->buffer);
	code->attributes.code.code_count = buffer_read_uint32(code->buffer);
	code->attributes.code.code       = &(code->buffer->data[code->buffer->next]);
	buffer_skip_bytes(code->buffer, code->attributes.code.code_count);
	code->attributes.code.handler_count = buffer_read_uint16(code->buffer);

	code->attributes.code.handlers   = \
		(JVExceptionHandler *)calloc(code->attributes.code.handler_count, sizeof(JVExceptionHandler));

	for(i = 0; i < code->attributes.code.handler_count; i++)
	{
		code->attributes.code.handlers[i].start_pc   = buffer_read_uint16(code->buffer);
		code->attributes.code.handlers[i].end_pc     = buffer_read_uint16(code->buffer);
		code->attributes.code.handlers[i].handler_pc = buffer_read_uint16(code->buffer);
		code->attributes.code.handlers[i].catch_type = buffer_read_uint16(code->buffer);
	}

	return OK;
}
